==> trend_etl-1.2.0.tar.gz: trend_etl-1.2.0/trend_etl/transform/discord_stats.py <==
from typing import Callable
from pyspark.sql import DataFrame, Column

from pyspark.sql.functions import col, date_add, now, ifnull, lit
from pyspark.sql.types import StringType, BinaryType, TimestampType

import trend_etl.schema.args as arguments
from trend_etl.functions.index import conv_timestamp

##
## conditions
##


def filter_trend_employees_cond(user_id_field: str) -> Column:
    return col(user_id_field).isin(arguments.discord_trend_employees) == False


def days_ago_cond(timestamp: Column, days_ago: int) -> Column:
    return ifnull(date_add(now(), -days_ago) <= timestamp, lit(False))


def greater_days_cond(timestamp: Column, days_ago: int) -> Column:
    return ifnull(date_add(now(), -days_ago) > timestamp, lit(False))


##
## stat functions
##


# total registered users: get all members for all guilds
def total_registered_users_number(guild_members_table: DataFrame) -> int:
    return guild_members_table.where(filter_trend_employees_cond("user.id")).count()


# total journeys: total # of journeys, specifically posts to the corresponding
#                 jounreys channel (posts are threads)
def total_journeys_number(
    channels_table: DataFrame, parent_channel_ids: list[str]
) -> int:
    # fetch threads (channels) that have the journeys channel as its parent
    journeys = channels_table.where(
        col("source_keys.parent_id").isin(parent_channel_ids)
    )

    # filter out trend employees
    filtered = journeys.where(filter_trend_employees_cond("owner_id"))

    return filtered.count()


# journeys conversion: number of registered users that created a journey
def journey_conversion_number(
    channels_table: DataFrame,
    guild_members_table: DataFrame,
    parent_channel_ids: list[str],
) -> int:
    # fetch threads (channels) that have the journeys channel as its parent
    journeys = channels_table.where(
        col("source_keys.parent_id").isin(parent_channel_ids)
    )

    # filter out trend employees
    filtered_journeys = journeys.where(filter_trend_employees_cond("owner_id"))
    filtered_registered_users = guild_members_table.where(
        filter_trend_employees_cond("user.id")
    )

    # get registered users that have created a journey
    conversions = filtered_registered_users.alias("member").join(
        filtered_journeys.alias("journeys"),
        col("journeys.owner_id") == col("member.user.id"),
        "left_semi",
    )

    return conversions.count()


# total statments: total # statments in all for all discord guilds (threads + messages)
def total_statements_number(threads_table: DataFrame, messages_table: DataFrame) -> int:
    # get count of all threads that are not trend employees
    threads_filtered = threads_table.where(filter_trend_employees_cond("owner_id"))
    threads_filtered_number = threads_filtered.count()

    # get a count all messages that are not trend employees
    messages = messages_table.where(filter_trend_employees_cond("author.id"))
    messages_number = messages.count()

    return threads_filtered_number + messages_number


# active users: # of users who have posted or replied in the last 30, 60, 90 days
def active_users_days_ago_number(
    threads_table: DataFrame, messages_table: DataFrame, days_ago: int
) -> int:
    thread_created = conv_timestamp("thread_metadata.create_timestamp")
    message_created = col("created_at")

    active_users = (
        active_users_for_threads(threads_table, days_ago_cond(thread_created, days_ago))
        .unionAll(
            active_users_for_messages(
                messages_table, days_ago_cond(message_created, days_ago)
            )
        )
        .drop_duplicates()  # NOTE: these functions don't perform de-duplication
        .count()
    )

    return active_users


# active audience: # of users who have posted, replied, or reacted to any message in the last 30 days
#                  this stat is essentially active users + "lurkers"
def active_audience_days_ago_number(
    threads_table: DataFrame,
    messages_table: DataFrame,
    reactions_table: DataFrame,
    days_ago: int,
) -> int:
    thread_created = conv_timestamp("thread_metadata.create_timestamp")
    message_created = col("created_at")

    active_audience = (
        active_users_for_threads(threads_table, days_ago_cond(thread_created, days_ago))
        .unionAll(
            active_users_for_messages(
                messages_table, days_ago_cond(message_created, days_ago)
            )
        )
        .unionAll(
            active_users_for_reactions(
                messages_table,
                reactions_table,
                days_ago_cond(message_created, days_ago),
            )
        )
        .drop_duplicates()  # NOTE: these functions don't perform de-duplication
        .count()
    )

    return active_audience


# Retention:
#   The number of users who interacted within <=`days_partition` day ago
#   on the server, who also have interacted on the server
#   >`days_partition` days ago
def retention_audience_number(
    threads_table: DataFrame,
    messages_table: DataFrame,
    reactions_table: DataFrame,
    days_partition: int,
) -> int:
    thread_created = conv_timestamp("thread_metadata.create_timestamp")
    message_created = col("created_at")

    # get users (dupes likely) that make up the active audience for
    # <=`days_partition` days ago
    active_audience_users_recent = (
        active_users_for_threads(
            threads_table, days_ago_cond(thread_created, days_partition)
        )
        .unionAll(
            active_users_for_messages(
                messages_table, days_ago_cond(message_created, days_partition)
            )
        )
        .unionAll(
            active_users_for_reactions(
                messages_table,
                reactions_table,
                days_ago_cond(message_created, days_partition),
            )
        )
    )

    # get users (dupes likely) that make up the active audience for
    # >`days_partition` days ago
    active_audience_users_former = (
        active_users_for_threads(
            threads_table, greater_days_cond(thread_created, days_partition)
        )
        .unionAll(
            active_users_for_messages(
                messages_table, greater_days_cond(message_created, days_partition)
            )
        )
        .unionAll(
            active_users_for_reactions(
                messages_table,
                reactions_table,
                greater_days_cond(message_created, days_partition),
            )
        )
    )

    audience_retention_users = active_audience_users_recent.intersect(
        active_audience_users_former
    )

    return audience_retention_users.count()


##
## stats helpers
##


# returns: DataFrame with single column `user_id`
def active_users_for_threads(
    threads_table: DataFrame, threads_date_cond: Column
) -> DataFrame:
    # get count of all threads that are not trend employees
    threads_filtered = threads_table.where(filter_trend_employees_cond("owner_id"))

    # get all threads that meet the date condition
    threads_days = threads_filtered.where(threads_date_cond)

    return threads_days.selectExpr("owner_id as user_id")


# returns: DataFrame with single column `user_id`
def active_users_for_messages(
    messages_table: DataFrame, messages_date_cond: Column
) -> DataFrame:
    # get count of all messages that are not trend employees
    messages_filtered = messages_table.where(filter_trend_employees_cond("author.id"))

    # get all messages that meet the date condition
    messages_days = messages_filtered.where(messages_date_cond)

    return messages_days.selectExpr("author.id as user_id")


# returns: DataFrame with single column `user_id`
def active_users_for_reactions(
    messages_table: DataFrame, reactions_table: DataFrame, messages_date_cond: Column
) -> DataFrame:
    # NOTE: we don't want to filter trend employee messages since a
    #       non- trend employee may react and that should count too.

    # get all messages that meet the date condition
    messages_days = messages_table.where(messages_date_cond)

    # get all the reactions for all messages
    messages_reactions = reactions_table.join(messages_days, ["doc_uuid"], "left_semi")

    # filter out reactions from trend employees
    reactions_filtered = messages_reactions.where(
        filter_trend_employees_cond("user.id")
    )

    return reactions_filtered.selectExpr("user.id as user_id")

==> trend_etl-1.4.2.tar.gz: trend_etl-1.4.2/trend_etl/transform/discord_stats.py <==
from typing import Callable
from pyspark.sql import DataFrame, Column

from pyspark.sql.functions import col, date_add, now, ifnull, lit
from pyspark.sql.types import StringType, BinaryType, TimestampType

import trend_etl.schema.args as arguments
from trend_etl.functions.index import conv_timestamp

##
## conditions
##


def filter_trend_employees_cond(user_id_field: str) -> Column:
    return col(user_id_field).isin(arguments.discord_trend_employees) == False


def days_ago_cond(timestamp: Column, days_ago: int) -> Column:
    return ifnull(date_add(now(), -days_ago) <= timestamp, lit(False))


def greater_days_cond(timestamp: Column, days_ago: int) -> Column:
    return ifnull(date_add(now(), -days_ago) > timestamp, lit(False))


##
## stat functions
##


# total registered users: get all members for all guilds
def total_registered_users_number(guild_members_table: DataFrame) -> int:
    return guild_members_table.where(filter_trend_employees_cond("user.id")).count()


# total journeys: total # of journeys, specifically posts to the corresponding
#                 jounreys channel (posts are threads)
def total_journeys_number(
    channels_table: DataFrame, parent_channel_ids: list[str]
) -> int:
    # fetch threads (channels) that have the journeys channel as its parent
    journeys = channels_table.where(
        col("source_keys.parent_id").isin(parent_channel_ids)
    )

    # filter out trend employees
    filtered = journeys.where(filter_trend_employees_cond("owner_id"))

    return filtered.count()


# journeys conversion: number of registered users that created a journey
def journey_conversion_number(
    channels_table: DataFrame,
    guild_members_table: DataFrame,
    parent_channel_ids: list[str],
) -> int:
    # fetch threads (channels) that have the journeys channel as its parent
    journeys = channels_table.where(
        col("source_keys.parent_id").isin(parent_channel_ids)
    )

    # filter out trend employees
    filtered_journeys = journeys.where(filter_trend_employees_cond("owner_id"))
    filtered_registered_users = guild_members_table.where(
        filter_trend_employees_cond("user.id")
    )

    # get registered users that have created a journey
    conversions = filtered_registered_users.alias("member").join(
        filtered_journeys.alias("journeys"),
        col("journeys.owner_id") == col("member.user.id"),
        "left_semi",
    )

    return conversions.count()


# total statments: total # statments in all for all discord guilds (threads + messages)
def total_statements_number(threads_table: DataFrame, messages_table: DataFrame) -> int:
    # get count of all threads that are not trend employees
    threads_filtered = threads_table.where(filter_trend_employees_cond("owner_id"))
    threads_filtered_number = threads_filtered.count()

    # get a count all messages that are not trend employees
    messages = messages_table.where(filter_trend_employees_cond("author.id"))
    messages_number = messages.count()

    return threads_filtered_number + messages_number


# active users: # of users who have posted or replied in at least `message_count_threshold`
#               times in the last 30, 60, 90 days
def active_users_days_ago_number(
    threads_table: DataFrame,
    messages_table: DataFrame,
    days_ago: int,
    message_count_threshold: int,
) -> int:
    thread_created = conv_timestamp("thread_metadata.create_timestamp")
    message_created = col("created_at")

    active_users = active_users_for_threads(
        threads_table, days_ago_cond(thread_created, days_ago)
    ).unionAll(
        active_users_for_messages(
            messages_table, days_ago_cond(message_created, days_ago)
        )
    )

    active_users_num = (
        active_users.groupBy("user_id")
        .count()
        .where(col("count") >= message_count_threshold)
        .count()
    )

    return active_users_num


# active audience: # of users who have posted or replied in at least `message_count_threshold`
#                  times in the last `days_ago` days PLUS # of users who have reacted at least
#                  `reaction_count_threshold` in the last `days_ago` days.
#                  Essentially Active Users + "lurkers"
def active_audience_days_ago_number(
    threads_table: DataFrame,
    messages_table: DataFrame,
    reactions_table: DataFrame,
    days_ago: int,
    message_count_threshold: int,
    reaction_count_threshold: int,
) -> int:
    thread_created = conv_timestamp("thread_metadata.create_timestamp")
    message_created = col("created_at")

    # Get all active users using the `days_ago` condition
    active_users = active_users_for_threads(
        threads_table, days_ago_cond(thread_created, days_ago)
    ).unionAll(
        active_users_for_messages(
            messages_table, days_ago_cond(message_created, days_ago)
        )
    )

    # Get users that reacted using the `days_ago` condition and,
    # filter out users already accounted for in `active_users`. In other words,
    # this collects all users that reacted that are not in the `active_users`
    # collection.
    active_users_reactions = active_users_for_reactions(
        messages_table,
        reactions_table,
        days_ago_cond(message_created, days_ago),
    ).join(active_users, "user_id", "left_anti")

    # active users + reactions = active audience
    # (uses the appropriate thresholds)

    active_users_num = (
        active_users.groupBy("user_id")
        .count()
        .where(col("count") >= message_count_threshold)
        .count()
    )

    active_users_reactions_num = (
        active_users_reactions.groupBy("user_id")
        .count()
        .where(col("count") >= reaction_count_threshold)
        .count()
    )

    return active_users_num + active_users_reactions_num


# Retention:
#   The number of users who interacted within <=`days_partition` day ago
#   on the server, who also have interacted on the server
#   >`days_partition` days ago
def retention_audience_number(
    threads_table: DataFrame,
    messages_table: DataFrame,
    reactions_table: DataFrame,
    days_partition: int,
) -> int:
    thread_created = conv_timestamp("thread_metadata.create_timestamp")
    message_created = col("created_at")

    # get users (dupes likely) that make up the active audience for
    # <=`days_partition` days ago
    active_audience_users_recent = (
        active_users_for_threads(
            threads_table, days_ago_cond(thread_created, days_partition)
        )
        .unionAll(
            active_users_for_messages(
                messages_table, days_ago_cond(message_created, days_partition)
            )
        )
        .unionAll(
            active_users_for_reactions(
                messages_table,
                reactions_table,
                days_ago_cond(message_created, days_partition),
            )
        )
    )

    # get users (dupes likely) that make up the active audience for
    # >`days_partition` days ago
    active_audience_users_former = (
        active_users_for_threads(
            threads_table, greater_days_cond(thread_created, days_partition)
        )
        .unionAll(
            active_users_for_messages(
                messages_table, greater_days_cond(message_created, days_partition)
            )
        )
        .unionAll(
            active_users_for_reactions(
                messages_table,
                reactions_table,
                greater_days_cond(message_created, days_partition),
            )
        )
    )

    audience_retention_users = active_audience_users_recent.intersect(
        active_audience_users_former
    )

    return audience_retention_users.count()


##
## stats helpers
##


# returns: DataFrame with single column `user_id`
def active_users_for_threads(
    threads_table: DataFrame, threads_date_cond: Column
) -> DataFrame:
    # get count of all threads that are not trend employees
    threads_filtered = threads_table.where(filter_trend_employees_cond("owner_id"))

    # get all threads that meet the date condition
    threads_days = threads_filtered.where(threads_date_cond)

    return threads_days.selectExpr("owner_id as user_id")


# returns: DataFrame with single column `user_id`
def active_users_for_messages(
    messages_table: DataFrame, messages_date_cond: Column
) -> DataFrame:
    # get count of all messages that are not trend employees
    messages_filtered = messages_table.where(filter_trend_employees_cond("author.id"))

    # get all messages that meet the date condition
    messages_days = messages_filtered.where(messages_date_cond)

    return messages_days.selectExpr("author.id as user_id")


# returns: DataFrame with single column `user_id`
def active_users_for_reactions(
    messages_table: DataFrame, reactions_table: DataFrame, messages_date_cond: Column
) -> DataFrame:
    # NOTE: we don't want to filter trend employee messages since a
    #       non- trend employee may react and that should count too.

    # get all messages that meet the date condition
    messages_days = messages_table.where(messages_date_cond)

    # get all the reactions for all messages
    messages_reactions = reactions_table.join(messages_days, ["doc_uuid"], "left_semi")

    # filter out reactions from trend employees
    reactions_filtered = messages_reactions.where(
        filter_trend_employees_cond("user.id")
    )

    return reactions_filtered.selectExpr("user.id as user_id")
